@@ -1,13 +1,14 @@
-from collections import namedtuple, OrderedDict
+from collections import namedtuple, OrderedDict, Iterable
 from copy import deepcopy
 from datasketch import MinHash, MinHashLSHForest
 from functools import partial
 from fuzzywuzzy import fuzz, process
 from heapq import heappop, heappush
+from multipledispatch import dispatch
 import numpy as np
 import warnings
 
-from htools.core import ngrams, tolist, identity, func_name
+from htools.core import ngrams, tolist, identity, func_name, listlike, select
 from htools.meta import add_docstring
 
 
@@ -904,6 +905,85 @@ class PriorityQueue:
         return f'{func_name(type(self))}({self._items})'
 
 
+class VocabDict(dict):
+    """Convenient interface for working with vocabularies where we want to map
+    from word -> index and from index -> word. Note that item deletion is not
+    supported, nor is changing the index of a key already in the dict.
+    The typical use case only involves creating the object once and leaving it.
+
+    Examples
+    --------
+    >>> w2i = {'the': 0, 'I': 1, 'door': 2, 'ocotopus': 3}
+    >>> vocab = VocabDict(w2i)
+    >>> vocab['door']
+    2
+    >>> vocab[2]
+    'door'
+    >>> vocab[[3, 0, 1]]
+    ['octopus', 'the', 'I']
+    >>> vocab['I', 2]
+    [1, 'door]
+    """
+
+    def __init__(self, w2i):
+        """
+        Parameters
+        ----------
+        w2i: Mapping[str, int]
+            Maps strings to numeric indices.
+            Ex: {'a': 0, 'the': 1, ... 'ostrich': 100}. Indices must be
+            consecutive and zero-indexed, though we don't force you to sort
+            this before calling the constructor.
+        """
+        if sorted(w2i.values()) != list(range(max(w2i.values()) + 1)):
+            raise ValueError(
+                'w2i indices must be consecutive and zero-indexed.')
+        # It's nice for the dict to be ordered appropriately, and we need to
+        # sort anyway to ensure i2w is in the correct order.
+        w2i = dict(sorted(w2i.items(), key=lambda x: x[1]))
+        super().__init__(w2i)
+        self.i2w = list(w2i)
+
+    @dispatch(int)
+    def __getitem__(self, i):
+        return self.i2w[i]
+
+    @dispatch(str)
+    def __getitem__(self, key):
+        return super().__getitem__(key)
+
+    @dispatch(Iterable)
+    def __getitem__(self, vals):
+        return [self[v] for v in vals]
+
+    def __setitem__(self, key, i):
+        if not isinstance(key, str):
+            raise TypeError(f'key must have type str, not {type(key)}.')
+        if not isinstance(i, int):
+            raise TypeError(f'i must have type int, not {type(i)}.')
+        if key in self:
+            raise ValueError(
+                f'Key {repr(key)} is already in {type(self).__name__}.')
+        if i != len(self):
+            raise ValueError(
+                f'Invalid value i={repr(i)}. Should be {len(self)} '
+                'because indices must be consecutive.')
+        super().__setitem__(key, i)
+        self.i2w.append(key)
+
+    def __delitem__(self, key):
+        raise RuntimeError(
+            f'{type(self).__name__} does not support item deletion.')
+
+    @classmethod
+    def fromkeys(cls, keys):
+        return cls({k: i for i, k in enumerate(keys)})
+
+    def setdefault(self, key, default):
+        raise RuntimeError(
+            f'setdefault method not supported for {type(self).__name__}.')
+
+
 class IndexedDict(OrderedDict):
     """OrderedDict that lets us use integer indices. The tradeoff is that we
     can no longer use integers as keys since that would make it ambiguous
@@ -948,3 +1028,78 @@ class LambdaDict(dict):
     def __missing__(self, key):
         self[key] = self.f(key)
         return self[key]
+
+
+class Results:
+    """Experimenting with ways to address a few shortcomings of Args and
+    DotDict, so there's some redundancy here.
+
+    1. IIRC Args is impossible to pickle, I think do some hacky naming choices
+    by me. Though I vaguely recall namedtuple itself may have similar problems
+    so that might have happened anyway.
+    2. DotDict addresses #1 (you do of course need to import htools to load
+    it, but I always have it imported anyway), but does not support tuple-style
+    unpacking.
+
+    But basically this is meant to be a simple way to return data from a
+    function, optionally pickle it, and conveniently load it again.
+    """
+
+    def __init__(self, **kwargs):
+        self._keys = list(kwargs)
+        self.__dict__.update(kwargs)
+
+    @classmethod
+    def from_dict(cls, data):
+        return cls(**data)
+
+    def from_dict(cls, data):
+        return cls(**data)
+
+    def __repr__(self):
+        arg_strs = ", ".join(f'{k}={v!r}' for k, v in self.items())
+        return f'{type(self).__name__}({arg_strs})'
+
+    def __iter__(self):
+        yield from self.values()
+
+    def __len__(self):
+        return len(self._keys)
+
+    def __getitem__(self, i:int):
+        """Notice this accepts an integer index and returns a value, NOT a key.
+        This is because I'd like to be able to unpack this like a tuple rather
+        than a dict.
+        """
+        keys = self._keys[i]
+        if listlike(keys):
+            return type(self)(**{key: getattr(self, key) for key in keys})
+        return getattr(self, keys)
+
+    def __getstate__(self):
+        return select(vars(self), drop=['keys'])
+
+    def __setstate__(self, data):
+        self._keys = []
+        for k, v in data.items():
+            self._keys.append(k)
+            self.__dict__[k] = v
+
+    def __eq__(self, o2):
+        return vars(self) == vars(o2)
+
+    def items(self):
+        return self._asdict().items()
+
+    def keys(self):
+        return self._asdict().keys()
+
+    def values(self):
+        return self._asdict().values()
+
+    def _asdict(self):
+        """Matches namedtuple interface. Also, we can't call dict directly
+        because __getitem__ works differently than a dict so we get an error.
+        """
+        return select(vars(self), keep=self._keys)
+
